from ml_gym.loss_functions.loss_functions import LPLoss, CrossEntropyLoss, BCEWithLogitsLoss, BCELoss, NLLLoss, Loss  # , LPLossScaled
from typing import Dict
from ml_gym.batching.batch_filters import BatchFilter


class LossFactory:

    @staticmethod
    def get_lp_loss(target_subscription_key: str, prediction_subscription_key: str, root: int = 1, exponent: int = 2,
                    class_selection_fun_params: Dict = None, average_batch_loss: bool = True, tag: str = "") -> Loss:
        sample_selection_fun = BatchFilter.get_class_selection_fun(
            **class_selection_fun_params) if class_selection_fun_params is not None else None
        return LPLoss(target_subscription_key, prediction_subscription_key, root, exponent, sample_selection_fun, tag, average_batch_loss)

    # @staticmethod
    # def get_scaled_lp_loss(target_subscription_key: str, prediction_subscription_key: str, root: int = 1, exponent: int = 2,
    #                        class_selection_fun_params: Dict = None, tag: str = "") -> Loss:
    #     sample_selection_fun = BatchFilter.get_class_selection_fun(**class_selection_fun_params) if class_selection_fun_params is not None else None
    #     return LPLossScaled(target_subscription_key, prediction_subscription_key, root, exponent, sample_selection_fun, tag)

    @staticmethod
    def get_cross_entropy_loss(target_subscription_key: str, prediction_subscription_key: str, tag: str = "") -> Loss:
        return CrossEntropyLoss(target_subscription_key=target_subscription_key,
                                prediction_subscription_key=prediction_subscription_key,
                                tag=tag)

    @staticmethod
    def get_nll_loss(target_subscription_key: str, prediction_subscription_key: str, tag: str = "") -> Loss:
        return NLLLoss(target_subscription_key=target_subscription_key,
                       prediction_subscription_key=prediction_subscription_key,
                       tag=tag)

    @staticmethod
    def get_bce_with_logits_loss(target_subscription_key: str, prediction_subscription_key: str, tag: str = "",
                                 average_batch_loss: bool = True, flatten_predictions: bool = False) -> Loss:
        return BCEWithLogitsLoss(target_subscription_key=target_subscription_key,
                                 prediction_subscription_key=prediction_subscription_key,
                                 average_batch_loss=average_batch_loss,
                                 tag=tag,
                                 flatten_predictions=flatten_predictions)

    @staticmethod
    def get_bce_loss(target_subscription_key: str, prediction_subscription_key: str, tag: str = "",
                     average_batch_loss: bool = True) -> Loss:
        return BCELoss(target_subscription_key=target_subscription_key,
                       prediction_subscription_key=prediction_subscription_key,
                       average_batch_loss=average_batch_loss,
                       tag=tag)



import torch
from torch import nn
from abc import ABC, abstractmethod
from ml_gym.batching.batch import InferenceResultBatch


class PredictPostProcessingIF(ABC):

    @abstractmethod
    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        raise NotImplementedError


class PredictPostProcessing(PredictPostProcessingIF):

    def __init__(self, postprocessing_impl: PredictPostProcessingIF):
        self.postprocessing_impl = postprocessing_impl

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        return self.postprocessing_impl.postprocess(result_batch)


# Postprocessing Implementations

class SoftmaxPostProcessorImpl(PredictPostProcessingIF):
    def __init__(self, prediction_subscription_key: str, prediction_publication_key: str):
        self.prediction_subscription_key = prediction_subscription_key
        self.prediction_publication_key = prediction_publication_key

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        predictions = nn.Softmax(dim=1)(result_batch.get_predictions(self.prediction_subscription_key))
        result_batch.add_predictions(key=self.prediction_publication_key, predictions=predictions)
        return result_batch


class ArgmaxPostProcessorImpl(PredictPostProcessingIF):
    def __init__(self, prediction_subscription_key: str, prediction_publication_key: str):
        self.prediction_subscription_key = prediction_subscription_key
        self.prediction_publication_key = prediction_publication_key

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        predictions = torch.argmax(result_batch.get_predictions(self.prediction_subscription_key), dim=1)
        result_batch.add_predictions(key=self.prediction_publication_key, predictions=predictions)
        return result_batch


class MaxOrMinPostProcessorImpl(PredictPostProcessingIF):
    def __init__(self, prediction_subscription_key: str, prediction_publication_key: str, agg_fun: str):
        self.prediction_subscription_key = prediction_subscription_key
        self.prediction_publication_key = prediction_publication_key
        if agg_fun == "max":
            self.agg_fun = torch.max
        elif agg_fun == "min":
            self.agg_fun = torch.min
        else:
            raise NotImplementedError("agg_fun={agg_fun} is not an applicable parameter")

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        predictions = self.agg_fun(result_batch.get_predictions(self.prediction_subscription_key), dim=1)[0]
        result_batch.add_predictions(key=self.prediction_publication_key, predictions=predictions)
        return result_batch


class BinarizationPostProcessorImpl(PredictPostProcessingIF):
    def __init__(self, prediction_subscription_key: str, prediction_publication_key: str, threshold=0.5):
        self.prediction_subscription_key = prediction_subscription_key
        self.prediction_publication_key = prediction_publication_key
        self.threshold = threshold

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        predictions = result_batch.get_predictions(self.prediction_subscription_key)
        binarized_outputs = torch.zeros_like(predictions).int()
        binarized_outputs[predictions > self.threshold] = 1
        result_batch.add_predictions(key=self.prediction_publication_key, predictions=binarized_outputs)
        return result_batch


class SigmoidalPostProcessorImpl(PredictPostProcessingIF):
    def __init__(self, prediction_subscription_key: str, prediction_publication_key: str, threshold=0.5):
        self.prediction_subscription_key = prediction_subscription_key
        self.prediction_publication_key = prediction_publication_key

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        predictions = result_batch.get_predictions(self.prediction_subscription_key)
        sigmoidal_predictions = torch.sigmoid(predictions)
        result_batch.add_predictions(key=self.prediction_publication_key, predictions=sigmoidal_predictions)
        return result_batch


class DummyPostProcessorImpl(PredictPostProcessingIF):

    def postprocess(self, result_batch: InferenceResultBatch) -> InferenceResultBatch:
        return result_batch


from abc import abstractmethod
from typing import Dict, List, Callable, Any
from ml_gym.loss_functions.loss_functions import Loss
from ml_gym.models.nn.net import NNModel
from ml_gym.data_handling.dataset_loader import DatasetLoader
import torch
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.gym.stateful_components import StatefulComponent
from ml_gym.optimizers.optimizer import OptimizerAdapter
import tqdm
from ml_gym.util.logger import ConsoleLogger
from ml_gym.gym.post_processing import PredictPostProcessingIF
from ml_gym.error_handling.exception import ModelAlreadyFullyTrainedError


class TrainComponent(StatefulComponent):
    def __init__(self, inference_component: InferenceComponent, post_processors: List[PredictPostProcessingIF],
                 loss_fun: Loss, show_progress: bool = False):
        self.loss_fun = loss_fun
        self.inference_component = inference_component
        self.post_processors = post_processors
        self.show_progress = show_progress
        self.logger = ConsoleLogger("logger_train_component")

    def train_batch(self, batch: DatasetBatch, model: NNModel, optimizer: OptimizerAdapter, device: torch.device):
        model.zero_grad()
        batch.to_device(device)
        loss = self.calc_loss(model, batch)
        loss.sum().backward()
        optimizer.step()

    def train_epoch(self, model: NNModel, optimizer: OptimizerAdapter, data_loader: DatasetLoader,
                    device: torch.device, epoch: int) -> NNModel:
        data_loader.device = device
        self.map_batches(fun=self.train_batch,
                         loader=data_loader,
                         fun_params={"device": device,
                                     "model": model,
                                     "optimizer": optimizer},
                         progress_info=f"Training {data_loader.dataset_name}  @epoch {epoch}")
        return model

    # def warm_up(self, model: NNModel, data_loader: DatasetLoader, device: torch.device):
    #     def init_loss(loss_fun: Loss, batch: InferenceResultBatch):
    #         if isinstance(loss_fun, LossWarmupMixin):
    #             loss_fun.warm_up(batch)
    #             loss_fun.finish_warmup()

    #     def check_if_initiable_component(loss_fun: Loss):
    #         return isinstance(loss_fun, LossWarmupMixin)

    #     if check_if_initiable_component(self.loss_fun):
    #         self.logger.log(LogLevel.INFO, "Running warmup...")
    #         with torch.no_grad():
    #             prediction_batches = self.map_batches(fun=self.forward_batch,
    #                                                   fun_params={"calculation_device": device, "model": model,
    #                                                               "result_device": torch.device("cpu")},
    #                                                   loader=data_loader,
    #                                                   show_progress=self.show_progress)
    #         # problem
    #         prediction_batch = InferenceResultBatch.combine(prediction_batches)
    #         init_loss(self.loss_fun, prediction_batch)
    #     else:
    #         self.logger.log(LogLevel.INFO, "Skipping training warmup. No special loss functions to be initialized.")

    def forward_batch(self, dataset_batch: DatasetBatch, model: NNModel, device: torch.device,) -> InferenceResultBatch:
        model = model.to(device)
        dataset_batch.to_device(device)
        inference_result_batch = self.inference_component.predict(model, dataset_batch, self.post_processors)
        return inference_result_batch

    def calc_loss(self, model: NNModel, batch: DatasetBatch) -> torch.Tensor:
        forward_batch = self.inference_component.predict(model, batch)
        loss = self.loss_fun(forward_batch)
        return loss

    @staticmethod
    def map_batches(fun: Callable[[DatasetBatch, NNModel], Any], loader: DatasetLoader,
                    fun_params: Dict[str, Any] = None, progress_info: str = None) -> List[InferenceResultBatch]:
        """
        Applies a function to each dataset_batch within a DatasetLoader
        """
        fun_params = fun_params if fun_params is not None else dict()
        if progress_info is not None:
            return [fun(dataset_batch, **fun_params) for dataset_batch in tqdm.tqdm(loader, desc=progress_info)]
        else:
            return [fun(dataset_batch, **fun_params) for dataset_batch in loader]


class TrainerIF(StatefulComponent):

    @abstractmethod
    def is_done(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def set_num_epochs(self, num_epochs: int):
        raise NotImplementedError

    @abstractmethod
    def set_current_epoch(self, epoch: int):
        raise NotImplementedError


class Trainer(TrainerIF):
    def __init__(self, train_component: TrainComponent, train_loader: DatasetLoader, verbose=False):
        self.train_component = train_component
        self.train_loader = train_loader
        self.verbose = verbose
        self.current_epoch = 1
        self.num_epochs = -1

    def is_done(self) -> bool:
        return self.current_epoch > self.num_epochs  # training starts at epoch 1, epoch 0 is just for evaluation, therefore >

    def set_num_epochs(self, num_epochs: int):
        self.num_epochs = num_epochs

    def set_current_epoch(self, epoch: int):
        self.current_epoch = epoch

    # def warm_up(self, model: NNModel, device: torch.device):
    #     self.train_component.warm_up(model, self.train_loader, device)

    def train_epoch(self, model: NNModel, optimizer: OptimizerAdapter, device: torch.device) -> NNModel:
        if self.current_epoch > self.num_epochs:
            raise ModelAlreadyFullyTrainedError(f"Model has been already trained for {self.current_epoch}/{self.num_epochs} epochs.")
        model = self.train_component.train_epoch(model, optimizer, self.train_loader, device, self.current_epoch)
        self.current_epoch += 1
        return model



from ml_gym.error_handling.exception import SamplerNotFoundError
from torch.utils.data import DataLoader
from torch.utils.data.sampler import RandomSampler, WeightedRandomSampler, Sampler, SequentialSampler
from typing import Callable, Dict, Any
from data_stack.dataset.iterator import InformedDatasetIteratorIF
from collections import Counter
import torch
from ml_gym.data_handling.postprocessors.collator import Collator
from enum import Enum


class DatasetLoaderFactory:

    @staticmethod
    def get_splitted_data_loaders_deprecated(dataset_splits: Dict[str, InformedDatasetIteratorIF], batch_size: int, collate_fn: Callable = None,
                                             weigthed_sampling_split_name: str = None, label_pos: int = 2, seeds: Dict[str, int] = None,
                                             drop_last: bool = False) -> Dict[str, "DatasetLoader"]:
        seeds = {} if seeds is None else seeds
        # NOTE: Weighting is only applied to the split specified by `weigthed_sampling_split_name`.
        data_loaders = {}
        for split_id, (split_name, dataset_split) in enumerate(dataset_splits.items()):
            seed = seeds[split_name] if split_name in seeds else None
            if split_name == weigthed_sampling_split_name:
                sampler = SamplerFactory.get_weighted_sampler(dataset_split, label_pos, seed)
            else:
                sampler = SamplerFactory.get_random_sampler(dataset_split, seed)

            data_loaders[split_name] = DatasetLoader(dataset_iterator=dataset_split,
                                                     batch_size=batch_size,
                                                     sampler=sampler,
                                                     collate_fn=collate_fn,
                                                     drop_last=drop_last)
        return data_loaders

    @staticmethod
    def get_splitted_data_loaders(dataset_splits: Dict[str, InformedDatasetIteratorIF], batch_size: int, collate_fn: Callable = None,
                                  drop_last: bool = False, sampling_strategies: Dict[str, Any] = None) -> Dict[str, "DatasetLoader"]:
        data_loaders = {}
        for split_name, dataset_split in dataset_splits.items():
            if split_name in sampling_strategies:
                config = sampling_strategies[split_name]
                strategy = SamplerFactory.SamplingStrategies[sampling_strategies[split_name]["strategy"]]
                config.pop("strategy")
                if strategy == SamplerFactory.SamplingStrategies.WEIGHTED_RANDOM:
                    sampler = SamplerFactory.get_weighted_sampler(dataset_split, **config)
                elif strategy == SamplerFactory.SamplingStrategies.RANDOM:
                    sampler = SamplerFactory.get_random_sampler(dataset_split, **config)
                elif strategy == SamplerFactory.SamplingStrategies.IN_ORDER:
                    sampler = SamplerFactory.get_sequential_sampler(dataset_split)
                else:
                    raise SamplerNotFoundError(f"Could not find sampler with key {strategy}")
            else:
                sampler = SamplerFactory.get_sequential_sampler(dataset_split)
            data_loaders[split_name] = DatasetLoader(dataset_iterator=dataset_split,
                                                     batch_size=batch_size,
                                                     sampler=sampler,
                                                     collate_fn=collate_fn,
                                                     drop_last=drop_last)
        return data_loaders


class SamplerFactory:

    class SamplingStrategies(Enum):
        RANDOM = "random"
        WEIGHTED_RANDOM = "weighted_random"
        IN_ORDER = "in_order"

    @staticmethod
    def get_weighted_sampler(dataset: InformedDatasetIteratorIF, label_pos: int = 2, seed: int = 0) -> Sampler:
        # """Returns a WeightedRandomSampler by counting the tags (sic!) over all samples.
        # Note, we don't count over targets as they might e.g., be one hot encoded.

        # Args:
        #     dataset (InformedDatasetIteratorIF): Iterator to calculate the sampler from

        # Returns:
        #     [WeightedRandomSampler]: Instance of WeightedRandomSampler.
        # """

        rnd_generator = torch.Generator().manual_seed(seed) if seed is not None else None
        # get the class weights
        target_counts = Counter([int(sample[label_pos]) for sample in dataset])  # uses generator expression
        target_tuples = [(k, v) for k, v in sorted(target_counts.items(), key=lambda item: item[0])]
        class_weights = {target_key: 1./target_count for target_key, target_count in target_tuples}
        sample_weights = [class_weights[sample[label_pos]] for sample in dataset]
        sampler = WeightedRandomSampler(weights=sample_weights, num_samples=len(dataset), generator=rnd_generator)
        return sampler

    @staticmethod
    def get_random_sampler(dataset: InformedDatasetIteratorIF, seed: int = 0) -> Sampler:
        rnd_generator = torch.Generator().manual_seed(seed) if seed is not None else None
        return RandomSampler(data_source=dataset, generator=rnd_generator)

    @staticmethod
    def get_sequential_sampler(dataset: InformedDatasetIteratorIF) -> Sampler:
        return SequentialSampler(data_source=dataset)


class DatasetLoader(DataLoader):
    def __init__(self, dataset_iterator: InformedDatasetIteratorIF, batch_size: int, sampler: Sampler,
                 collate_fn: Collator = None, drop_last: bool = False):
        super().__init__(dataset=dataset_iterator, sampler=sampler, batch_size=batch_size, collate_fn=collate_fn, drop_last=drop_last)

    @property
    def dataset_name(self) -> str:
        return self.dataset.dataset_meta.dataset_name

    @property
    def device(self) -> torch.device:
        return self.collate_fn.device

    @device.setter
    def device(self, d: torch.device):
        if self.collate_fn is not None:
            self.collate_fn.device = d



from torch.multiprocessing import Queue
import time
import tqdm
import torch
from typing import List
from ml_gym.multiprocessing.job import JobType, Job
from ml_gym.multiprocessing.worker import WorkerProcessWrapper
from ml_gym.util.logger import QueuedLogging
from ml_gym.util.logger import LogLevel, QLogger


class Pool:
    def __init__(self, num_processes: int, devices: List[torch.device], max_jobs_per_process: int = 1):
        self.num_processes = num_processes
        self._job_count = 0
        self.job_q = Queue()
        self.done_q = Queue()
        self.devices = devices
        self.worker_processes = []
        self.max_jobs_per_process = max_jobs_per_process
        self.logger: QLogger = QueuedLogging.get_qlogger("logger_pool")
        self.logger.log(LogLevel.INFO, f"Initialized to run jobs on: {self.devices}")

    def add_job(self, job: Job):
        self.job_q.put(job)
        self._job_count += 1

    def add_jobs(self, jobs: List[Job]):
        self.logger.log(LogLevel.INFO, "Filling up job queue ... ")
        for job in tqdm.tqdm(jobs):
            self.job_q.put(job)
        self._job_count += len(jobs)

    @property
    def job_count(self):
        return self._job_count

    def run(self):
        jobs_done = 0
        # we have to add the termination jobs at the end of the queue such that the processes stop working and don't get stuck in jobs_q.get()
        termination_jobs = [Job(job_id=-1, fun=None, param_dict=None, job_type=JobType.TERMINATE) for _ in
                            range(self.num_processes)]
        self.add_jobs(termination_jobs)
        # create and start worker processes
        self.logger.log(LogLevel.INFO, f"Creating {self.num_processes} worker processes...")
        for process_id in tqdm.tqdm(range(self.num_processes)):
            self.create_or_replace_process(process_id, self.max_jobs_per_process)
        self.logger.log(LogLevel.INFO, f"Starting {self.num_processes} worker processes...")
        for p in tqdm.tqdm(self.worker_processes):
            p.start()
        # wait until all jobs are done
        for _ in range(self._job_count):
            done_job = self.done_q.get()
            jobs_done += 1
            if done_job.error is not None:
                self.logger.log(LogLevel.FATAL,
                                f"FATAL! Job {done_job.job_id} crashed while being executed by process {done_job.executing_process_id} on {done_job.device} after {int(done_job.finishing_time - done_job.starting_time)} seconds with error {done_job.error}")
                self.logger.log(LogLevel.FATAL, f"Error: {done_job.stacktrace}")
            else:
                self.logger.log(LogLevel.INFO,
                                f"Job {done_job.job_id} was successfully executed by process {done_job.executing_process_id} on {done_job.device} within {int(done_job.finishing_time - done_job.starting_time)} seconds.")
                self.logger.log(LogLevel.INFO, f"Progress: {int(jobs_done / self._job_count * 100)}%")
            if done_job.job_type == JobType.CALC:
                self.worker_processes[done_job.executing_process_id].recreate_process_if_done()

    def create_or_replace_process(self, process_id: int, num_jobs_to_perform: int):
        process = WorkerProcessWrapper(process_id=process_id,
                                       num_jobs_to_perform=num_jobs_to_perform,
                                       device=self.devices[process_id % len(self.devices)],
                                       job_q=self.job_q,
                                       done_q=self.done_q)
        if len(self.worker_processes) == process.process_id:
            self.logger.log(LogLevel.DEBUG, f"Adding process {process_id}.")
            self.worker_processes.append(process)
        elif len(self.worker_processes) > process.process_id:
            self.worker_processes[process_id] = process
            self.logger.log(LogLevel.DEBUG, f"Replaced process {process_id}.")
        else:
            raise Exception("There is something wrong with the process id!")

    if __name__ == "__main__":
        def do_sleep(seconds: int, device: torch.device):
            time.sleep(seconds)
            # print("Device: " + device)
            # raise Exception("bla")

        # pool = Pool(3, max_jobs_per_process=1)
        # pool.add_jobs(jobs)
        # pool.run()



from data_stack.dataset.iterator import DatasetIteratorIF
from ml_gym.data_handling.postprocessors.postprocessor import PostProcessorIf
from typing import List


class PostProcessedDatasetIterator(DatasetIteratorIF):

    def __init__(self, dataset_iterator: DatasetIteratorIF, post_processor: PostProcessorIf):
        self._dataset_iterator = dataset_iterator
        self._post_processor = post_processor

    def __len__(self):
        return len(self._dataset_iterator)

    def __getitem__(self, index: int):
        return self._post_processor.postprocess(self._dataset_iterator[index])

    @property
    def underlying_iterators(self) -> List[DatasetIteratorIF]:
        return [self._dataset_iterator]



from torch.optim.optimizer import Optimizer
from typing import Dict, Any, Type
from copy import deepcopy
from ml_gym.error_handling.exception import OptimizerNotInitializedError


class OptimizerAdapter(object):

    def __init__(self, optimizer_class: Type[Optimizer], optimizer_params: Dict = None):
        self._optimizer_class = optimizer_class
        self._optimizer_params = optimizer_params if optimizer_params is not None else {}

        self._optimizer: Optimizer = None
        self._state_dict = None

    def register_model_params(self, model_params: Dict, restore_state: bool=True):
        model_params_list = model_params.values()
        if not restore_state:
            self._optimizer = self._optimizer_class(**self._optimizer_params, params=model_params_list)
            return

        # since we instantiate a new optimizer when we register a model, we have to save and restore the optimizer state
        if self._optimizer is not None:
            state_dict = self.state_dict()
            self._optimizer = self._optimizer_class(**self._optimizer_params, params=model_params_list)
            self.load_state_dict(state_dict)
        else:
            self._optimizer = self._optimizer_class(**self._optimizer_params, params=model_params_list)
            if self._state_dict is not None:
                self._optimizer.load_state_dict(self._state_dict)
                self._state_dict = None

    def __getstate__(self):
        if self._optimizer is not None:
            return self._optimizer.__getstate__()
        else:
            return {}

    def __setstate__(self, state):
        if not hasattr(self, '_optimizer'):
            self._optimizer = None
        if self._optimizer is not None:
            self._optimizer.__setstate__(state)

    def __repr__(self):
        return self._optimizer.__repr__()

    def _hook_for_profile(self):
        self._optimizer._hook_for_profile()

    def state_dict(self) -> Dict[str, Any]:
        if self._optimizer is not None:
            return self._optimizer.state_dict()
        elif self._state_dict is not None:
            return self._state_dict
        else:
            raise OptimizerNotInitializedError("Cannot access state_dict, because internal optimizer was not instantiated.")


    def load_state_dict(self, state_dict: Dict):
        if self._optimizer is not None:
            self._optimizer.load_state_dict(state_dict)
        else:  # here we store the state_dict until we instantiate the optimizer
            self._state_dict = state_dict

    def zero_grad(self, set_to_none: bool = False):
        if self._optimizer is None:
            raise OptimizerNotInitializedError("Internal optimizer was not instantiated. Has a model been registered for this optimizer?")
        self._optimizer.zero_grad(set_to_none)

    def step(self, closure=None):
        if self._optimizer is None:
            raise OptimizerNotInitializedError("Internal optimizer was not instantiated. Has a model been registered for this optimizer?")
        self._optimizer.step(closure)

    def add_param_group(self, param_group):
        if self._optimizer is None:
            raise OptimizerNotInitializedError("Internal optimizer was not instantiated. Has a model been registered for this optimizer?")
        self._optimizer.add_param_group(param_group)

    @property
    def param_groups(self):
        return self._optimizer.param_groups

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result



from torch.multiprocessing import Process, Queue
import time
import torch
import traceback
from ml_gym.multiprocessing.job import Job, JobType
from ml_gym.util.logger import MLgymLoggerIF, LogLevel, QueuedLogging


class WorkerProcess(Process):
    def __init__(self, process_id: int, num_jobs_to_perform: int, job_q: Queue, done_q: Queue, device: torch.device, logger: MLgymLoggerIF):
        super(WorkerProcess, self).__init__(target=self.work, args=(job_q, done_q, num_jobs_to_perform, device, logger))
        self.process_id = process_id

    def work(self, job_q: Queue, done_q: Queue, num_jobs_to_perform: int, device: torch.device, logger: MLgymLoggerIF):

        logger.log(LogLevel.INFO, f"Process {self.process_id} started working.")
        jobs_done_count = 0
        for job in iter(job_q.get, None):  # https://stackoverflow.com/a/21157892
            job.device = device
            job.executing_process_id = self.process_id
            logger.log(LogLevel.INFO, f"Process {job.executing_process_id} started job {job.job_id} on {job.device}.")
            job.starting_time = time.time()
            if job.job_type == JobType.CALC:
                self._do_calc(job)
            job.finishing_time = time.time()
            job.done = True
            jobs_done_count += 1
            if job.job_type == JobType.TERMINATE or num_jobs_to_perform == jobs_done_count:
                logger.log(LogLevel.DEBUG, f"Process {self.process_id} terminated.")
                done_q.put(job)
                break

    def _do_calc(self, job: Job):
        try:
            job.execute()
        except Exception as e:
            job.error = str(e)
            job.stacktrace = traceback.format_exc()


class WorkerProcessWrapper:
    def __init__(self, process_id: int, num_jobs_to_perform: int, device: torch.device, job_q: Queue, done_q: Queue):
        self.logger = QueuedLogging.get_qlogger(f"logger_process_{process_id}", )
        self.jobs_done_count = 0
        self.device = device
        self.num_jobs_to_perform = num_jobs_to_perform
        self.process_id = process_id
        self.job_q = job_q
        self.done_q = done_q
        self.process = WorkerProcess(process_id, num_jobs_to_perform, job_q, done_q, device, self.logger)

    def recreate_process_if_done(self):
        self.jobs_done_count += 1
        if self.num_jobs_to_perform == self.jobs_done_count:
            self.logger.log(LogLevel.DEBUG, f"Recreating process {self.process_id}.")
            self.process = WorkerProcess(self.process_id, self.num_jobs_to_perform,
                                         self.job_q, self.done_q, self.device, self.logger)
            self.process.start()
            self.logger.log(LogLevel.DEBUG, f"Recreated process {self.process_id}.")
            self.jobs_done_count = 0

    def get_process_id(self) -> int:
        return self.process.process_id

    def start(self):
        return self.process.start()



from dataclasses import dataclass, field
from abc import abstractmethod
from typing import List, Callable
import torch


@dataclass
class Collator(Callable):

    device: torch.device = field(default_factory=lambda: torch.device("cpu"))

    @abstractmethod
    def __call__(self, batch: List[torch.Tensor]):
        """ Takes a batch and collates into a proper TrainBatch.
        :param batch
        :return:
        """
        raise NotImplementedError



class DatasetNotFoundError(Exception):
    """Exception raised when a dataset could not be found."""
    pass


class DatasetOutOfBoundsError(Exception):
    """Exception raised when an index >= len(Dataset) is used."""
    pass


class ModelNotTrainedError(Exception):
    """Exception raised when model state is requested that is training dependent"""
    pass


class ModelAlreadyFullyTrainedError(Exception):
    """Exception raised when model is fully trained but training for another epoch was initiated."""
    pass


class TrainingStateCorruptError(Exception):
    """Raised e.g., when the number of checkpointed models does not match the number of evaluations in the metrics.json"""


class DatasetFileCorruptError(Exception):
    """Thrown when integrity checks indicate that a given file is corrupt."""
    pass


class LossFunctionInitError(Exception):
    """Exception raised when a loss function cannot be initiliazed"""
    pass


class BatchStateError(Exception):
    """Exception raised when a Batch object does not contain the requested data"""
    pass


class ExperimentInfoMissing(Exception):
    """Exception raised when experiment info object is missing"""
    pass


class ComponentConstructionError(Exception):
    """Exception raised when an error occurred during component construction"""
    pass


class DependentComponentNotFoundError(Exception):
    """Exception raised when a component required by a parent component could not be found"""
    pass


class SingletonAlreadyInstantiatedError(Exception):
    """Exception raised when trying to get another instance from a singleton class"""
    pass


class InjectMappingNotFoundError(Exception):
    """Exception raised when Injector cannot inject given placehold when it was not defined in the mapping."""
    pass


class ValidationModeNotValidError(Exception):
    """Exception when an invalid validation mode was selected."""
    pass


class InvalidTensorFormatError(Exception):
    """Raised when a torch tensor has the wrong format for subsequent computation"""


class OptimizerNotInitializedError(Exception):
    """Raised when we want to run an operation on an optimizer, which was not instantiated, yet."""
    pass


class SamplerNotFoundError(Exception):
    """Raised when the sampler implemenation for a given datsetloader was not found."""
    pass



from enum import Enum
import torch
from typing import Callable, Dict


class JobType(Enum):
    CALC = 1
    TERMINATE = 2


class Job:
    def __init__(self, job_id: int, fun: Callable, param_dict: Dict, job_type: JobType = JobType.CALC):
        self.job_id = job_id
        self.job_type = job_type
        self.fun = fun
        self.param_dict = param_dict

        self.done = False
        self.starting_time = -1
        self.finishing_time = -1
        self.executing_process_id = -1
        self.error = None
        self.stacktrace = None
        self._device: torch.device = None

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, value: torch.device):
        self._device = value

    def execute(self):
        self.param_dict["device"] = self._device
        return self.fun(**self.param_dict)



from torch.optim import Optimizer, SGD, Adam, Adadelta
from typing import Dict
from ml_gym.optimizers.optimizer import OptimizerAdapter


class OptimizerFactory:
    optimizer_map: Dict[str, Optimizer] = {
        "SGD": SGD,
        "ADAM": Adam,
        "ADADELTA": Adadelta
    }

    @classmethod
    def get_optimizer(cls, optimizer_key: str, params: Dict) -> OptimizerAdapter:
        optimizer_class = cls.optimizer_map[optimizer_key]
        return OptimizerAdapter(optimizer_class=optimizer_class, optimizer_params=params)



import torch
from abc import abstractmethod, ABC
from typing import Dict, List, Any, Callable, Union
from ml_gym.error_handling.exception import BatchStateError
import copy
from functools import partial


class TorchDeviceMixin(ABC):

    @staticmethod
    def _dict_tensor_to_device(d: Dict[str, Any], device: torch.device) -> Dict[str, Any]:
        partial_fun = partial(torch.Tensor.to, device=device)
        return TorchDeviceMixin.traverse_apply(ds=d, apply_fun=partial_fun)
        # return {k: TorchDeviceMixin._dict_tensor_to_device(v, device) if not isinstance(v, torch.Tensor) else v.to(device) for k, v in d.items()}

    @staticmethod
    def _detach_dict_tensor(d: Dict[str, Any]) -> Dict[str, Any]:
        partial_fun = partial(torch.Tensor.detach)
        return TorchDeviceMixin.traverse_apply(ds=d, apply_fun=partial_fun)
        # return {k: TorchDeviceMixin._detach_dict_tensor(v) if not isinstance(v, torch.Tensor) else v.detach() for k, v in d.items()}

    @staticmethod
    def traverse_apply(ds: Union[Dict, List, torch.Tensor], apply_fun: Callable[[torch.Tensor], torch.Tensor]) -> Union[Dict, List, torch.Tensor]:
        if isinstance(ds, dict):
            return {k: TorchDeviceMixin.traverse_apply(d, apply_fun) for k, d in ds.items()}
        elif isinstance(ds, list):
            return [TorchDeviceMixin.traverse_apply(d, apply_fun) for d in ds]
        return apply_fun(ds)

    @abstractmethod
    def get_device(self) -> torch.device:
        raise NotImplementedError

    @abstractmethod
    def to_device(self, device: torch.device):
        raise NotImplementedError

    @abstractmethod
    def detach(self):
        raise NotImplementedError

    @abstractmethod
    def to_cpu(self):
        raise NotImplementedError


class Batch(ABC):
    """Abstract class that defines the necessary methods any `Batch` implementation needs to implement.
    """

    @staticmethod
    def combine(batches: List['Batch']):
        batch_combined = batches[0].__class__.combine_impl(batches)
        return batch_combined

    @staticmethod
    def _copy_tensor_dict(d: Dict[Any, torch.Tensor]):
        return {k: v.detach().clone() for k, v in d.items()}

    @staticmethod
    def _combine_tensor_dicts(tensor_dicts: List[Dict[Any, torch.Tensor]]) -> Dict[Any, torch.Tensor]:
        combined_tensor_dict = {}
        for key in tensor_dicts[0].keys():
            if isinstance(tensor_dicts[0][key], dict):
                sub_tensor_dicts = [d[key] for d in tensor_dicts]
                combined_tensor_dict[key] = Batch._combine_tensor_dicts(sub_tensor_dicts)
            else:
                combined_tensor_dict[key] = torch.cat([d[key] for d in tensor_dicts])
        return combined_tensor_dict

    @staticmethod
    @abstractmethod
    def combine_pair(b_1: 'Batch', b_2: 'Batch') -> 'Batch':
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def combine_impl(batches: List['Batch']) -> 'Batch':
        raise NotImplementedError


class DatasetBatch(Batch, TorchDeviceMixin):
    """A batch of samples and its targets and tags. Used to batch train a model."""

    def __init__(self, samples: Union[torch.Tensor, Dict], targets: Dict[str, torch.Tensor], tags: torch.Tensor = None,
                 samples_require_grad: bool = False):
        self._samples = samples
        self._samples.requires_grad = samples_require_grad
        self._targets = targets
        self._tags = tags if tags is not None else torch.Tensor()

    @property
    def samples(self) -> Union[torch.Tensor, Dict]:
        return self._samples

    @property
    def targets(self) -> Dict[str, torch.Tensor]:
        return self._targets

    @property
    def tags(self) -> torch.Tensor:
        return self._tags

    @property
    def samples_require_grad(self) -> bool:
        return self._samples.requires_grad

    @samples_require_grad.setter
    def samples_require_grad(self, value: bool):
        self._samples.requires_grad = value

    def detach(self):
        self._targets = {k: v.detach() for k, v in self._targets.items()}
        self._tags = self._tags.detach()
        self._samples = self._samples.detach()

    def to_device(self, device: torch.device):
        self._samples = self._samples.to(device)
        self._targets = {k: v.to(device) for k, v in self._targets.items()}
        self._tags = self._tags.to(device)

    def to_cpu(self):
        self.to_device(device=torch.device("cpu"))

    def get_device(self) -> torch.device:
        return self._samples.device

    @staticmethod
    def combine_impl(batches: List['DatasetBatch']) -> 'DatasetBatch':
        tags = torch.cat([batch.tags for batch in batches])
        samples = torch.cat([batch.samples for batch in batches])
        targets = Batch._combine_tensor_dicts([batch.targets for batch in batches])
        return DatasetBatch(targets=targets, samples=samples, tags=tags)

    def __len__(self) -> int:
        return len(self._samples)

    def __deepcopy__(self, memo) -> 'DatasetBatch':
        samples_ = self.samples.detach().clone()
        targets_ = self._copy_tensor_dict(self.targets)
        tags_ = self.tags.detach().clone()
        return DatasetBatch(samples=samples_, targets=targets_, tags=tags_)

    @staticmethod
    def combine_pair(b_1: 'DatasetBatch', b_2: 'DatasetBatch') -> 'DatasetBatch':
        b_1 = copy.deepcopy(b_1)
        b_2 = copy.deepcopy(b_2)
        tags = torch.cat([b_1.tags, b_2.tags])
        samples = torch.cat([b_1.samples, b_2.samples])
        targets = Batch._combine_tensor_dicts([b_1.targets, b_2.targets])
        return DatasetBatch(targets=targets, samples=samples, tags=tags)


class InferenceResultBatch(Batch, TorchDeviceMixin):
    """ Stores targets and predictions of an entire batch.
    """

    def __init__(self, targets: Dict[str, torch.Tensor] = None, predictions: Dict[str, torch.Tensor] = None, tags: torch.Tensor = None):
        self._targets = targets if targets is not None else {}
        self._tags = tags
        self._predictions = predictions if predictions is not None else {}

    def to_cpu(self):
        self.to_device(device=torch.device("cpu"))

    def get_device(self) -> torch.device:
        return self._tags.device

    def to_device(self, device: torch.device):
        self._predictions = TorchDeviceMixin._dict_tensor_to_device(self._predictions, device)
        self._targets = {k: v.to(device) for k, v in self._targets.items()}
        self._tags = self._tags.to(device)

    def detach(self):
        self._targets = {k: v.detach() for k, v in self._targets.items()}
        self._tags = self._tags.detach()
        self._predictions = TorchDeviceMixin._detach_dict_tensor(self._predictions)

    @property
    def predictions(self) -> Dict[str, torch.Tensor]:
        return self._predictions

    def add_predictions(self, key: str, predictions: torch.Tensor):
        self._predictions[key] = predictions

    def get_predictions(self, key: str) -> torch.Tensor:
        if key not in self._predictions:
            raise BatchStateError(f"Key {key} not present in predictions!")
        return self._predictions[key]

    def drop_predictions(self, keys: List[str]):
        for key in keys:
            del self._predictions[key]

    def drop_targets(self, keys: List[str]):
        for key in keys:
            del self._targets[key]

    def get_targets(self, key: str) -> torch.Tensor:
        if key not in self._targets:
            raise BatchStateError(f"Key {key} not present in targets!")
        return self._targets[key]

    def add_targets(self, key: str, targets: torch.Tensor):
        self._targets[key] = targets

    @property
    def targets(self) -> Dict[str, torch.Tensor]:
        return self._targets

    @property
    def tags(self) -> torch.Tensor:
        return self._tags

    def __len__(self) -> int:
        return len(self._tags)

    def __deepcopy__(self, memo) -> 'InferenceResultBatch':
        predictions_ = self._copy_tensor_dict(self.predictions)
        targets_ = self._copy_tensor_dict(self.targets)
        tags_ = self.tags.detach().clone()
        return InferenceResultBatch(predictions=predictions_, targets=targets_, tags=tags_)

    def split_results(self, target_keys: List[str], predictions_keys: List[Union[str, List]], device: torch.device):
        def _filter_predictions(predictions_keys: List[str], predictions: Dict, filtered_predictions: Dict):
            p_key = predictions_keys[0]
            if p_key == "*":
                p_keys = list(predictions.keys())
            else:
                p_keys = [p_key]
            if len(predictions_keys) == 1:
                for p_key in p_keys:
                    filtered_predictions[p_key] = predictions[p_key]
            else:
                for p_key in p_keys:
                    if p_key not in filtered_predictions:
                        filtered_predictions[p_key] = {}
                    _filter_predictions(predictions_keys[1:], predictions[p_key], filtered_predictions[p_key])

        predictions_keys_list = [[p_key] if isinstance(p_key, str) else p_key for p_key in predictions_keys]
        filtered_targets = {key: self._targets[key].to(device) for key in target_keys if key in self._targets}

        filtered_predictions = {}
        for p_keys in predictions_keys_list: 
            _filter_predictions(p_keys, self.predictions, filtered_predictions)
 
        filtered_predictions = TorchDeviceMixin._dict_tensor_to_device(filtered_predictions, device)
        tags = self.tags.to(device)
        return InferenceResultBatch(targets=filtered_targets, predictions=filtered_predictions, tags=tags)

    @staticmethod
    def combine_pair(b_1: 'InferenceResultBatch', b_2: 'InferenceResultBatch') -> 'InferenceResultBatch':
        # tags = torch.cat([b_1.tags, b_2.tags])
        tags = b_1.tags + b_2.tags
        predictions = Batch._combine_tensor_dicts([b_1.predictions, b_2.predictions])
        targets = Batch._combine_tensor_dicts([b_1.targets, b_2.targets])
        return InferenceResultBatch(targets=targets, predictions=predictions, tags=tags)

    @staticmethod
    def combine_impl(batches: List['InferenceResultBatch']) -> 'InferenceResultBatch':
        tags = [tag for batch in batches for tag in batch.tags]
        predictions = Batch._combine_tensor_dicts([batch.predictions for batch in batches])
        targets = Batch._combine_tensor_dicts([batch.targets for batch in batches])
        return InferenceResultBatch(targets=targets, predictions=predictions, tags=tags)


class EvaluationBatchResult(Batch):
    """Data class for storing the results of a single or multiple batches. Also entire epoch results are stored in here.
    """

    def __init__(self, losses: Dict[str, List[float]], metrics: Dict[str, List[float]], dataset_name: str,
                 split_name: str):
        self._losses = losses
        self._metrics = metrics
        self._dataset_name = dataset_name
        self._split_name = split_name

    @property
    def losses(self) -> Dict[str, List[float]]:
        return self._losses

    @property
    def metrics(self) -> Dict[str, List[float]]:
        return self._metrics

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @property
    def split_name(self) -> str:
        return self._split_name

    def aggregate(self, fun: Callable[[List], List] = None):
        if fun is None:
            def fun(e): return [sum(e)]
        self._losses = {k: fun(v) for k, v in self._losses.items()}
        self._metrics = {k: fun(v) for k, v in self._metrics.items()}

    @staticmethod
    def combine_pair(b_1: 'EvaluationBatchResult', b_2: 'EvaluationBatchResult') -> 'EvaluationBatchResult':
        raise NotImplementedError

    @staticmethod
    def combine_impl(batches: List['EvaluationBatchResult']) -> 'EvaluationBatchResult':
        raise NotImplementedError

    def __str__(self) -> str:
        eval_str = f"Evaluation result on {self._dataset_name} ({self._dataset_split}):"
        eval_str += "\n\nlosses: " + "\n\t".join([f"{k}: {v}" for k, v in self._losses.items()])
        eval_str += "\n\nmetrics: " + "\n\t".join([f"{k}: {v}" for k, v in self._metrics.items()])
        eval_str += "\n==============================================="
        return eval_str

    def to_dict(self) -> Dict[str, Any]:
        results = {str(k): v for k, v in self.metrics.items()}
        for k, v in self.losses.items():
            k = str(k)
            if str(k) in results:
                raise BatchStateError("Metrics keys and losses keys must be distinct")
            else:
                results[k] = v
        return results



from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.models.nn.net import NNModel
import json
import os
import torch
from data_stack.dataset.iterator import DatasetIteratorIF
from typing import Dict, List, Type, Callable
from copy import deepcopy
from ml_gym.data_handling.dataset_loader import DatasetLoader, DatasetLoaderFactory
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.util.grid_search import GridSearch
from ml_gym.validation.validator_factory import ValidatorFactory
from ml_gym.io.config_parser import YAMLConfigLoader
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch
from ml_gym.gym.predict_postprocessing_component import PredictPostprocessingComponent
from ml_gym.gym.post_processing import PredictPostProcessingIF
import tqdm
from ml_gym.gym.jobs import AbstractGymJob
from data_stack.dataset.iterator import InformedDatasetIteratorIF


class ExportedModel:
    def __init__(self, model: NNModel, post_processors: List[PredictPostProcessingIF], model_path: str = None, device: torch.device = None):
        self.model = model
        self.post_processors = post_processors
        self.model_path = model_path
        self._device = device if device is not None else torch.device("cpu")
        self.model.to(self._device)

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, d: torch.device):
        self._device = d
        self.model.to(self._device)

    def predict_tensor(self, sample_tensor: torch.Tensor, targets: torch.Tensor = None, tags: torch.Tensor = None, no_grad: bool = True):
        sample_tensor = sample_tensor.to(self._device)
        if no_grad:
            with torch.no_grad():
                forward_result = self.model.forward(sample_tensor)
        else:
            forward_result = self.model.forward(sample_tensor)
        result_batch = InferenceResultBatch(targets=targets, tags=tags, predictions=forward_result)
        result_batch = PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)
        return result_batch

    def predict_dataset_batch(self, batch: DatasetBatch, no_grad: bool = True) -> InferenceResultBatch:
        batch.to_device(self._device)
        if no_grad:
            with torch.no_grad():
                forward_result = self.model.forward(batch.samples)
        else:
            forward_result = self.model.forward(batch.samples)

        result_batch = InferenceResultBatch(targets=deepcopy(batch.targets), tags=deepcopy(batch.tags), predictions=forward_result)
        result_batch = PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)
        result_batch.to_cpu()
        return result_batch

    def predict_dataset_iterator(self, dataset_iterator: InformedDatasetIteratorIF,
                                 batch_size: int, collate_fn: Callable, no_grad: bool = True) -> InferenceResultBatch:
        split_key = "dataset_split"
        sampling_strategies = {split_key: {"strategy": "IN_ORDER"}}
        dataset_loader = DatasetLoaderFactory.get_splitted_data_loaders({split_key: dataset_iterator}, batch_size=batch_size,
                                                                        sampling_strategies=sampling_strategies,
                                                                        collate_fn=collate_fn)[split_key]
        irb = self.predict_data_loader(dataset_loader=dataset_loader, no_grad=no_grad)
        return irb

    def predict_data_loader(self, dataset_loader: DatasetLoader, no_grad: bool = True) -> InferenceResultBatch:
        dataset_loader.device = self._device
        result_batches = [self.predict_dataset_batch(batch, no_grad) for batch in tqdm.tqdm(dataset_loader, desc="Batches processed:")]
        return InferenceResultBatch.combine(result_batches)

    @staticmethod
    def from_model_and_preprocessors(model: NNModel, post_processors: List[PredictPostProcessingIF], model_path: str, device: torch.device = None) -> "ExportedModel":
        return ExportedModel(model, post_processors, model_path, device=device)


class ComponentLoader:

    @staticmethod
    def get_trained_exported_model(components: List, experiment_path: str, model_id: int,
                                   split_name: str, device: torch.device = None) -> ExportedModel:
        trained_model = ComponentLoader.get_trained_model(components, experiment_path, model_id, device)
        post_processors = components["eval_component"].post_processors[split_name]
        model_path = os.path.join(experiment_path, f"checkpoints/model_{model_id}.pt")
        exported_model = ExportedModel.from_model_and_preprocessors(trained_model, post_processors, model_path, device)
        return exported_model

    @staticmethod
    def get_components(experiment_path: str, blueprint_type: Type[BluePrint], component_names: List[str]):
        config_path = os.path.join(experiment_path, "config.json")

        with open(config_path, "r") as fd:
            config = json.load(fd)
        components = blueprint_type.construct_components(config=config, component_names=component_names)
        return components

    @staticmethod
    def get_components_from_grid_search(gs_path: str, blueprint_type: Type[BluePrint], component_names: List[str], gs_id: int = 0):
        run_id_to_config_dict = {str(run_id): config for run_id, config in enumerate(GridSearch.create_gs_configs_from_path(gs_path))}
        experiment_config = run_id_to_config_dict[f"{gs_id}"]
        components = blueprint_type.construct_components(config=experiment_config, component_names=component_names)
        return components

    @staticmethod
    def get_components_from_nested_cv(gs_path: str, cv_path: str, blueprint_type: Type[BluePrint], component_names: List[str]):
        gs_config = YAMLConfigLoader.load(gs_path)
        cv_config = YAMLConfigLoader.load(cv_path)
        nested_cv = ValidatorFactory.get_nested_cv(gs_config=gs_config,
                                                   cv_config=cv_config,
                                                   grid_search_id="bla",
                                                   blue_print_type=blueprint_type)
        blue_print = nested_cv.create_blue_prints(blueprint_type, AbstractGymJob.Type.STANDARD, gs_config, 1, dashify_logging_path="")[0]
        components = blueprint_type.construct_components(
            config=blue_print.config, component_names=component_names, external_injection=blue_print.external_injection)
        return components

    @staticmethod
    def get_trained_model(components: List, experiment_path: str, model_id: int, device: torch.device = None) -> NNModel:
        model_state_dict_path = os.path.join(experiment_path, f"checkpoints/model_{model_id}.pt")
        model = deepcopy(components["model"])
        # load model
        model_state = torch.load(model_state_dict_path, map_location=device)
        model.load_state_dict(model_state)
        model = model.to(device)
        return model

    @staticmethod
    def get_datasets(components: List) -> Dict[str, DatasetIteratorIF]:
        dataset_keys = list(
            components["evaluator"].eval_component.dataset_loaders.keys())
        datasets_dict = {
            key: components["evaluator"].eval_component.dataset_loaders[key].dataset for key in dataset_keys}
        return datasets_dict

    @staticmethod
    def get_dataloaders(components: List) -> Dict[str, DatasetLoader]:
        return components["evaluator"].eval_component.dataset_loaders

    @staticmethod
    def get_inference_component(components: List) -> InferenceComponent:
        return components["evaluator"].eval_component.inference_component



from typing import Dict, Any, Tuple, List, Type
from data_stack.dataset.iterator import DatasetIteratorIF
from ml_gym.blueprints.blue_prints import create_blueprint
from ml_gym.gym.gym import Gym
from ml_gym.gym.jobs import AbstractGymJob
from data_stack.dataset.splitter import SplitterFactory
from ml_gym.validation.validator import ValidatorIF
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.blueprints.component_factory import Injector
from ml_gym.util.grid_search import GridSearch


class NestedCV(ValidatorIF):
    def __init__(self, dataset_iterator: DatasetIteratorIF, num_outer_loop_folds: int,
                 num_inner_loop_folds: int, inner_stratification: bool, outer_stratification: bool,
                 target_pos: int, shuffle: bool, grid_search_id: str, seed: int, re_eval: bool = False,
                 keep_interim_results: bool = True):
        self.num_outer_loop_folds = num_outer_loop_folds
        self.num_inner_loop_folds = num_inner_loop_folds
        self.inner_stratification = inner_stratification
        self.outer_stratification = outer_stratification
        self.dataset_iterator = dataset_iterator
        self.seed = seed
        self.grid_search_id = grid_search_id
        self.target_pos = target_pos
        self.shuffle = shuffle
        self.re_eval = re_eval
        self.keep_interim_results = keep_interim_results

    def _get_fold_indices(self) -> Tuple[List[int]]:
        splitter = SplitterFactory.get_nested_cv_splitter(num_inner_loop_folds=self.num_inner_loop_folds,
                                                          num_outer_loop_folds=self.num_outer_loop_folds,
                                                          inner_stratification=self.inner_stratification,
                                                          outer_stratification=self.outer_stratification,
                                                          target_pos=self.target_pos,
                                                          shuffle=self.shuffle,
                                                          seed=self.seed)
        indices = splitter.get_indices(dataset_iterator=self.dataset_iterator)
        return indices

    @staticmethod
    def _create_outer_folds_splits(outer_folds_indices: Tuple[List[int]]) -> List[Dict[str, Any]]:
        # outer folds
        # [outer_fold_1_indices, outer_fold2_indices ...]
        splits = []
        for outer_fold_id, test_fold_indices in enumerate(outer_folds_indices):
            # create train fold
            train_fold_indices = []
            for i, fold in enumerate(outer_folds_indices):
                if i != outer_fold_id:
                    train_fold_indices = train_fold_indices + fold

            split = {
                "id_outer_test_fold_id": outer_fold_id,
                "id_inner_test_fold_id": -1,
                "id_split_indices": {
                    "train": train_fold_indices,
                    "test": test_fold_indices
                }
            }
            splits.append(split)
        return splits

    @staticmethod
    def _create_inner_folds_splits(inner_folds_indices: Tuple[List[int]]) -> List[Dict[str, Any]]:
        splits = []
        # inner folds
        # [[outer_fold_1_inner_fold_1_indices, outer_fold_1_inner_fold_2_indices, ...], ...]
        for outer_fold_id, inner_folds in enumerate(inner_folds_indices):
            for inner_fold_id, test_fold_indices in enumerate(inner_folds):
                # calc train fold indices
                train_fold_indices = []
                for i, fold in enumerate(inner_folds):
                    if i != inner_fold_id:
                        train_fold_indices = train_fold_indices + fold
                split = {
                    "id_outer_test_fold_id": outer_fold_id,
                    "id_inner_test_fold_id": inner_fold_id,
                    "id_split_indices": {
                        "train": train_fold_indices,
                        "test": test_fold_indices
                    }
                }
                splits.append(split)
        return splits

    def create_blue_prints(self, blue_print_type: Type[BluePrint], job_type: AbstractGymJob.Type, gs_config: Dict[str, Any], num_epochs: int,
                           dashify_logging_path: str) -> List[Type[BluePrint]]:

        run_id_to_config_dict = {run_id: config for run_id, config in enumerate(GridSearch.create_gs_from_config_dict(gs_config))}

        outer_fold_indices, inner_folds_indices = self._get_fold_indices()

        outer_splits = NestedCV._create_outer_folds_splits(outer_fold_indices)
        inner_splits = NestedCV._create_inner_folds_splits(inner_folds_indices)
        splits = outer_splits + inner_splits

        blueprints = []
        experiment_id = 0
        for split in splits:
            for config_id, experiment_config in run_id_to_config_dict.items():
                external_injection = {"id_experiment_id": experiment_id,
                                      "id_hyper_paramater_combination_id": config_id,
                                      **split}
                injector = Injector(mapping=external_injection)
                experiment_config_injected = injector.inject_pass(component_parameters=experiment_config)
                bp = create_blueprint(blue_print_class=blue_print_type,
                                      run_mode=AbstractGymJob.Mode.TRAIN if not self.re_eval else AbstractGymJob.Mode.EVAL,
                                      job_type=job_type,
                                      experiment_config=experiment_config_injected,
                                      dashify_logging_path=dashify_logging_path,
                                      num_epochs=num_epochs,
                                      grid_search_id=self.grid_search_id,
                                      experiment_id=experiment_id)
                blueprints.append(bp)
                experiment_id = experiment_id + 1
        return blueprints

    def run(self, blue_print_type: Type[BluePrint], gym: Gym, gs_config: Dict[str, Any], num_epochs: int, dashify_logging_path: str):
        job_type = AbstractGymJob.Type.STANDARD if self.keep_interim_results else AbstractGymJob.Type.LITE
        blueprints = self.create_blue_prints(blue_print_type=blue_print_type,
                                             gs_config=gs_config,
                                             dashify_logging_path=dashify_logging_path,
                                             num_epochs=num_epochs,
                                             job_type=job_type)
        gym.add_blue_prints(blueprints)
        gym.run(parallel=True)
